# -*- coding: utf-8 -*-

# Copyright (c) 2014, Joshua Stone
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from .request import get_comments, url_comments
from .comment import Comment
from .helpers import search_comments_fields, api_key, join_params, set_limit, validate_filter

__all__ = [
  "Comments"
]

class Comments(object):
  """
  Comments() is the interface for interacting with Derpibooru's API similar on Search().

  All properties are read-only, and every method returns a new instance of
  Comments() to avoid mutating state in ongoing search queries. This makes object
  interactions predictable as well as making versioning of searches relatively
  easy.
  """
  def __init__(self, key="", q=set(), limit=50, filter_id="",
               author="", body="", created_at="", comment_id="", image_id="",
               my=None, user_id="", per_page=25, page=1,
               url_domain="https://derpibooru.org", proxies={}):
    """
    By default initializes an instance of Comments with the parameters to get
    the first 25 comments on Derpibooru's comments activity page.
    """
    self.proxies = proxies
    self.url_domain = url_domain
    self._params = {
      "key": api_key(key),
      "q": search_comments_fields(q, author=author, body=body,
                                  created_at=created_at, comment_id=comment_id,
                                  image_id=image_id, my=my, user_id=user_id),
      "filter_id": validate_filter(filter_id),
      "per_page": set_limit(per_page),
      "page": set_limit(page)
    }
    self._limit = set_limit(limit)
    self._search = get_comments(self._params, self._limit, url_domain=self.url_domain, proxies=self.proxies)
  
  def __iter__(self):
    """
    Make Comments() iterable so that new search results can be lazily generated
    for performance reasons.
    """
    return self

  @property
  def parameters(self):
    """
    Returns a list of available parameters; useful for passing state to new
    instances of Comments().
    """
    return self._params

  @property
  def url(self):
    """
    Returns a search URL built on set parameters. Example based on default
    parameters:

    https://derpibooru.org/comments?qc=%2A
    """
    return url_comments(self._params, url_domain=self.url_domain)

  def key(self, key=""):
    """
    Takes a user's API key string which applies content settings. API keys can
    be found at <https://derpibooru.org/registration/edit>.
    """
    params = join_params(self.parameters, {"key": key,
                                           "limit": self._limit,
                                           "url_domain": self.url_domain,
                                           "proxies": self.proxies}
                        )

    return self.__class__(**params)

  def query(self, *q):
    """
    Takes one or more strings for searching by query.
    """
    params = join_params(self.parameters, {"q": q,
                                           "limit": self._limit,
                                           "url_domain": self.url_domain,
                                           "proxies": self.proxies}
                        )

    return self.__class__(**params)

  def limit(self, limit):
    """
    Set absolute limit on number of images to return, or set to None to return
    as many results as needed; default 50 posts. This limit on app-level.
    """
    params = join_params(self.parameters, {"limit": limit, 
                                           "url_domain": self.url_domain,
                                           "proxies": self.proxies})

    return self.__class__(**params)

  def filter(self, filter_id=""):
    """
    Takes a filter's ID to be used in the current search context. Filter IDs can
    be found at <https://derpibooru.org/filters/> by inspecting the URL parameters.
    
    If no filter is provided, the user's current filter will be used.
    """
    params = join_params(self.parameters, {"filter_id": validate_filter(filter_id),
                                           "limit": self._limit,
                                           "url_domain": self.url_domain,
                                           "proxies": self.proxies}
                        )

    return self.__class__(**params)

  def set_field(self, field, string):
    """
    Common function for set values in fields. When string starts with "-", appends tag "-field:value".
    """
    if string:
       if f"{string}".startswith("-"):
          query = self.query_remove(f"{field}:{string[1:]}").query_append(f"-{field}:{string[1:]}")
       else:
          query = self.query_remove(f"-{field}:{string}").query_append(f"{field}:{string}")
    else:
       exists_tags = set(filter(lambda tag: True
                                            if f"{tag}".startswith(f"field:")
                                               or f"{tag}".startswith(f"-{field}:")
                                            else False,
                                q)
                        )
       query = self.query_remove(exists_tags)
    return query

  def author(self, string=None):
    """
    Set the author of comment. Anonymous authors will never match this term.
    """
    return self.set_field("author", string)

  def created_at(self, string=None):
    """
    Set the creation time of comment in ISO 8601 format.
    """
    return self.set_field("created_at", string)

  def comment_id(self, string=None):
    """
    Set the numeric surrogate key for comment.
    """
    return self.set_field("id", string)

  def image_id(self, string=None):
    """
    Set the numeric surrogate key for the image comment belongs to.
    """
    return self.set_field("image_id", string)

  def my(self, my_value=None):
    """
    my:comments matches comments you have posted if you are signed in. 
    """
    if not self._params["key"] or my_value is None:
       return self.set_field("my", None)
    elif my_value:
       return self.set_field("my", "comments")
    else:
       return self.set_field("my", "-comments")

  def user_id(self, string=None):
    """
    Matches comments with the specified user_id. Anonymous users will never match this term.
    """
    return self.set_field("user_id", string)

  def query_append(self,*q):
     """
     Adds query to current search.
     """
     query = self._params['q'].union(q)
     params = join_params(self.parameters, {"q": query,
                                           "limit": self._limit,
                                           "url_domain": self.url_domain,
                                           "proxies": self.proxies}
                         )

     return self.__class__(**params)

  def query_remove(self,*q):
     """
     Removes query from current search.
     """
     query = self._params['q'].difference(q)
     params = join_params(self.parameters, {"q": query,
                                           "limit": self._limit,
                                           "url_domain": self.url_domain,
                                           "proxies": self.proxies}
                         )

     return self.__class__(**params)

  def get_page(self,page):
    """
    Set returned page of search.
    """
    params = join_params(self.parameters, {"page": set_limit(page),
                                           "limit": self._limit,
                                           "url_domain": self.url_domain,
                                           "proxies": self.proxies}
                        )

    return self.__class__(**params)

  def per_page(self,limit):
    """
    Set absolute limit on number of comments to get, or set to None to return
    defaulting 25 posts; max 50 posts. This limit on API-level.
    """
    params = join_params(self.parameters, {"per_page": set_limit(limit),
                                           "limit": self._limit,
                                           "url_domain": self.url_domain,
                                           "proxies": self.proxies}
                        )

    return self.__class__(**params)

  def __next__(self):
    """
    Returns a result wrapped in a new instance of Comment().
    """
    return Comment(next(self._search), url_domain=self.url_domain, proxies=self.proxies)



# -*- coding: utf-8 -*-

# Copyright (c) 2014, Joshua Stone
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from .request import post_image
from .tags import Tags
from .image import Image
from .search import Search
from .helpers import join_params, tags, api_key, validate_url, validate_description

__all__ = [
  "PostImage",
  "ImageError", "TagsError", "LengthError", "PostError", "DuplicateError"
]

class PostImage(object):
  """
  This class is a helper for submits a new image by using API.
  Abuse will result in a ban.
  You must provide the direct link to the image in the url parameter.
  """
  def __init__(self, key="", image_url="", description="", 
               tag_input=set(), source_url="",
               url_domain="https://derpibooru.org", proxies={}):

    self.proxies = proxies
    self.url_domain = url_domain
    self._params = {
      "key": api_key(key),
      "image_url": validate_url(image_url),
      "description": validate_description(description),
      "tag_input": tags(tag_input),
      "source_url": validate_url(source_url)
    }

  def __str__(self):
    params = ", ".join(f"{param}='{self.parameters[param]}'" for param in self.parameters)
    return f"PostImage({params})"

  @property
  def parameters(self):
    """
    Returns a list of available parameters; useful for passing state to new
    instances of PostImage().
    """
    return self._params

  def key(self, key=""):
    """
    Takes a user's API key string which applies content settings. API keys can
    be found at <https://derpibooru.org/registration/edit>.
    """
    params = join_params(self.parameters, {"key": key,
                                           "url_domain": self.url_domain,
                                           "proxies": self.proxies}
                        )

    return self.__class__(**params)

  def image_url(self, image_url=""):
    """
    Direct link to the image.
    """
    params = join_params(self.parameters, {"image_url": image_url,
                                           "url_domain": self.url_domain,
                                           "proxies": self.proxies}
                        )

    return self.__class__(**params)

  def description(self, description=""):
    """
    Describe this image in plain words - this should generally be info about
    the image that doesn't belong in the tags or source.
    This field supports simple html.
    Can be 50000 bytes max.
    """
    params = join_params(self.parameters, {"description": description,
                                           "url_domain": self.url_domain,
                                           "proxies": self.proxies}
                        )

    return self.__class__(**params)

  def tag_input(self, *tags):
    """
    Describe with 3+ tags, including ratings and applicable artist tags
    """
    params = join_params(self.parameters, {"tag_input": tags,
                                           "url_domain": self.url_domain,
                                           "proxies": self.proxies}
                        )

    return self.__class__(**params)

  def source_url(self, source_url=""):
    """
    Describe with 3+ tags, including ratings and applicable artist tags
    """
    params = join_params(self.parameters, {"source_url": source_url,
                                           "url_domain": self.url_domain,
                                           "proxies": self.proxies}
                        )

    return self.__class__(**params)

  def post(self,new_tags=False,dnp=False,dupe=False):
    """
    Post new image and return Image() object or raise error.
    Set "new_tags"==True for accept adding new tags.
    Set "dnp"=True for accept posting image from DNP artist.
    Set "dupe"=True for ignore duplicate checking.
    """
    if self.parameters['key'] and self.parameters['image_url']:
      self.validate_tags()
      if not new_tags:
        fix_tags = self.check_tags()[1]
        if fix_tags:
          raise TagsError(self.parameters['tag_input'], 
                          reason="this tags wasn't used before " \
                                f"({', '.join(f'{tag} -> {fix_tags[tag]}' if fix_tags[tag] else tag for tag in fix_tags)})")
      if not dnp:
        dnp_tags = self.dnp()
        if dnp_tags:
          br = '\n'
          dnp_reasons = br.join(f"{tag_name} : [\n" \
                                f"{br.join(dnp_entry['conditions'] for dnp_entry in dnp_tags[tag_name]) : >{len(tag_name)}}\n]" \
                                for tag_name in sorted(dnp_tags))
          raise TagsError(self.parameters['tag_input'], 
                          reason=f"DNP\n{dnp_reasons}")
      if not dupe:
        duplicates = self.check_duplicate()
        if duplicates:
          raise DuplicateError(self.parameters, duplicates)
      response = post_image(**self.parameters, url_domain=self.url_domain, proxies=self.proxies)
      if "errors" in response:
        raise ImageError(response["errors"])
      else:
        return Image(response["image"], key=self.parameters['key'],
                     url_domain=self.url_domain, proxies=self.proxies)
    else:
      raise PostError(self.parameters)

  def validate_tags(self):
    """
    Preposting tag validation will raise TagsError() if tag_input is incorrect.
    tag_input require:
    - 3+ tags;
    - at least one rating tag, and for non-safe images can have one of each non-safe type;
    - not blacklisted (you can found blacklist at <https://github.com/derpibooru/philomena/blob/master/config/tag.json>)
    """
    tag_input = self.parameters['tag_input']
    safe_rating = {"safe"}
    sexual_ratings = {"suggestive", "questionable", "explicit"}
    horror_ratings = {"semi-grimdark", "grimdark"}
    gross_rating = {"grotesque"}
    rating_tags = safe_rating | sexual_ratings | horror_ratings | sexual_ratings
    blacklist = {"tagme", "tag me", "not tagged", "no tag", "notag",
                 "notags", "upvotes galore", "downvotes galore", "wall of faves",
                 "drama in the comments", "drama in comments", "tag needed",
                 "paywall", "cringeworthy", "solo oc", "tag your shit"}
    if len(tag_input) < 3:
      reason = "must contain at least 3 tags"
    elif tag_input & blacklist: 
      reason = f"contains forbidden tag ({', '.join(sorted(tag_input & blacklist))})'"
    elif not (tag_input & rating_tags):
      reason = "must contain at least one rating tag"
    elif (safe_rating & tag_input) and (tag_input & (rating_tags-safe_rating)):
      reason = "may not contain any other rating if safe " \
              f"({', '.join(sorted(tag_input & (rating_tags-safe_rating)))})"
    elif len(sexual_ratings & tag_input) > 1:
      reason = "may contain at most one sexual rating " \
              f"({', '.join(sorted(sexual_ratings & tag_input))})"
    elif len(horror_ratings & tag_input) > 1:
      reason = "may contain at most one grim rating " \
              f"({', '.join(sorted(horror_ratings & tag_input))})"
    else:
      reason = ""
    if reason:
      raise TagsError(tag_input, reason)
    return True

  def check_tags(self):
    """
    Returns dict() with image count as value for every tag in tag_input.
    So, if tag has 0 image count, it means that tag does not exist on Derpibooru.
    """
    query = ' || '.join(f"name:{tag}" for tag in self.parameters['tag_input'])
    tags = Tags(q=(query,), per_page=50, limit=len(self.parameters['tag_input']),
                url_domain=self.url_domain, proxies=self.proxies)
    tags_count = {tag.name: (tag.allias_parent().images if tag.allias_parent() else tag.images) 
                  for tag in tags}
    num_tags = {tag_name:(tags_count[tag_name] if tag_name in tags_count else 0) 
                for tag_name in self.parameters['tag_input']}
    fix_tags = {}
    for tag_name in num_tags:
      if num_tags[tag_name]==0:
        ft = list(Tags(q=(f"name:{tag_name}~1.0",), per_page=50, limit=1,
                       url_domain=self.url_domain, proxies=self.proxies))
        if ft:
          fix_tags[tag_name] = ft[0].name
        else:
          fix_tags[tag_name] = None
    return num_tags, fix_tags

  def dnp(self):
    """
    Returns dict() of tags artists who have a restriction on the uploading
    of their artwork. Values are conditions for upload and empty condition
    means that only artist can upload.
    """
    query = ' || '.join(f"name:{tag}" for tag in self.parameters['tag_input'])
    tags = Tags(q=(query,), per_page=50, limit=len(self.parameters['tag_input']),
                url_domain=self.url_domain, proxies=self.proxies)
    dnp_tags = {tag.name:tag.dnp_entries for tag in tags if tag.dnp_entries}
    return dnp_tags

  def check_duplicate(self):
    """
    Return duplicates of image or False if image is unique.
    """
    reverse = tuple(Search(reverse_url=self.parameters['image_url'],
                          distance=0.25,
                          url_domain=self.url_domain, proxies=self.proxies))
    if reverse:
      return reverse
    return False

  def tags_append(self,*tags):
    """
    Adds tags to current tag_input.
    """
    new_tags = self.parameters['tag_input'].union(tags)
    params = join_params(self.parameters, {"tag_input": new_tags,
                                           "url_domain": self.url_domain,
                                           "proxies": self.proxies}
                        )

    return self.__class__(**params)

  def tags_remove(self,*tags):
    """
    Removes tags from current tag_input.
    """
    new_tags = self.parameters['tag_input'].difference(tags)
    params = join_params(self.parameters, {"tag_input": new_tags,
                                           "url_domain": self.url_domain,
                                           "proxies": self.proxies}
                        )

    return self.__class__(**params)

class ImageError(Exception):
  """
  Trying to upload image returns response with errors.
  """
  def __init__(self, data):
    self.data = data

  def __str__(self):
    data = '\n'.join(f'{item} : {self.data[item]}' for item in self.data)
    reasons = []
    if "image_format" in self.data:
      reasons.append("unsupported format")
    if "image_height" in self.data:
      reasons.append("probably too big height (32767 px max)")
    if "image_width" in self.data:
      reasons.append("probably too big width (32767 px max)")
    if "image_size" in self.data:
      reasons.append("probably too big size (26 214 400 max)")
    if "image_mime_type" in self.data:
      reasons.append('probably unsupported mime ' \
                     '(should be One of "image/gif", "image/jpeg", ' \
                     '"image/png", "image/svg+xml", "video/webm")')
    if "image_orig_sha512_hash" in self.data:
      reasons.append(f"image already exists in the database")
    if "tag_input" in self.data:
      reasons.append(f"errors with the tag metadata")

    return f"{', '.join(reasons)}\n{data}"

class TagsError(Exception):
  """
  Tags are incorrect error.
  """
  def __init__(self, tag_input, reason=""):
    self.tag_input = tag_input

    if reason:
       self.reason = reason
    else:
       safe_rating = {"safe"}
       sexual_ratings = {"suggestive", "questionable", "explicit"}
       horror_ratings = {"semi-grimdark", "grimdark"}
       gross_rating = {"grotesque"}
       rating_tags = rating_tags | sexual_ratings | horror_ratings | sexual_ratings
       blacklist = {"tagme", "tag me", "not tagged", "no tag", "notag",
                    "notags", "upvotes galore", "downvotes galore", "wall of faves",
                    "drama in the comments", "drama in comments", "tag needed",
                    "paywall", "cringeworthy", "solo oc", "tag your shit"}
       if len(tag_input) < 3:
         self.reason = "must contain at least 3 tags"
       elif tag_input & blacklist: 
         self.reason = f"contains forbidden tag ({', '.join(sorted(tag_input & blacklist))})'"
       elif not (tag_input & rating_tags):
         self.reason = "must contain at least one rating tag"
       elif (safe_rating & tag_input) and (tag_input & (rating_tags-safe_rating)):
         self.reason = "may not contain any other rating if safe " \
                      f"({', '.join(sorted(tag_input & (rating_tags-safe_rating)))})"
       elif len(sexual_ratings & tag_input) > 1:
         self.reason = "may contain at most one sexual rating " \
                      f"({', '.join(sorted(sexual_ratings & tag_input))})"
       elif len(horror_ratings & tag_input) > 1:
         self.reason = "may contain at most one grim rating " \
                      f"({', '.join(sorted(horror_ratings & tag_input))})"

  def __str__(self):
    return f"{self.reason}"

class LengthError(Exception):
  """
  Too long string error.
  """
  def __init__(self, string):
    self.string = string

  def __str__(self):
    message = f"'{self.string[:20]}' has " \
              f"{len(self.string.encode('utf-8'))} bytes when max is 50000"
    return message

class PostError(Exception):
  """
  PostImage() has empty required fields error.
  """
  def __init__(self, parameters):
    self.parameters = parameters

  def __str__(self):
    if self.parameters['key']:
      return "should include image_url"
    elif self.parameters['image_url']:
      return "should include key"
    else:
      return "should include key and image_url"

class DuplicateError(Exception):
  """
  Image already exists error.
  """
  def __init__(self, parameters, duplicates):
    self.parameters = parameters
    self.duplicates = duplicates

  def __str__(self):
    message = f"<{self.parameters['image_url']}> already exists " \
              f"({', '.join(f'{image.id}' for image in self.duplicates)})"
    return message


# -*- coding: utf-8 -*-

# Copyright (c) 2014, Joshua Stone
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from .request import get_galleries
from .search import Search
from .image import Image

__all__ = [
  "Gallery"
]

class Gallery(object):
  """
  This class provides a thin wrapper around JSON data, mapping each value to
  its own property. Once instantiated the data is immutable so as to reflect
  the stateless nature of a REST API.
  """
  def __init__(self, data, gallery_id=None, search_params={},
               url_domain="https://derpibooru.org", proxies={}):
    self.proxies = proxies
    self.url_domain = url_domain
    self._params = search_params

    if data is None and gallery_id:
      search_params['q'] = (f"id:{gallery_id}",)
      self._data = data = next(get_galleries(search_params, limit=1,
                                             url_domain=url_domain, proxies=proxies))
    else:
      self._data = data

    for field, body in data.items():
      if not hasattr(self, field):
        setattr(self, field, body) 

  def __str__(self):
    return f"Gallery({self.id})"
       
  @property
  def url(self):
    return f"{self.url_domain}/galleries/{self.id}"

  @property
  def data(self):
    return self._data

  def update(self):
    data = next(get_galleries(self._params, limit=1,
                              url_domain=self.url_domain, proxies=self.proxies))

    if data:
      self._data = data

  @property
  def thumbnail(self):
    return Image(None, image_id=self.thumbnail_id,
                 search_params=self._params,
                 url_domain=self.url_domain, proxies=self.proxies)
  
  def images(self, sf="created_at", sd="desc", limit=50,
             faves="", upvotes="", uploads="", watched="",
             filter_id="", per_page=25, page=1):
    return Search(key=self._params["key"] if "key" in self._params else "", 
                  q=(f"gallery_id:{self.id}",), sf=sf, sd=sd, limit=limit,
                  faves=faves, upvotes=upvotes, uploads=uploads, watched=watched,
                  filter_id=filter_id, per_page=per_page, page=page,
                  url_domain=self.url_domain, proxies=self.proxies)


#!/usr/bin/env python

from setuptools import setup
from setuptools import find_packages

setup(
  name = "DerPyBooruPhi",
  description = "Python bindings for Philomena-based boorus' API like Ponybooru, Derpibooru, etc.",
  url = "https://github.com/Atronar/DerPyBooruPhi",
  version = "0.10.3",
  author = "ATroN",
  author_email = "master.atron@gmail.com",
  license = "Simplified BSD License",
  platforms = ["any"],
  packages = find_packages(),
  python_requires='>=3.6',
  install_requires = ["requests"],
  include_package_data = True,
  #download_url = "https://github.com/joshua-stone/DerPyBooru/tarball/0.7.2",
  classifiers = [
    "Development Status :: 4 - Beta",
    "Intended Audience :: Developers",
    "Operating System :: OS Independent",
    "License :: OSI Approved :: BSD License",
    "Topic :: Software Development :: Libraries :: Python Modules",
    "Programming Language :: Python :: 3"
  ]
)



# -*- coding: utf-8 -*-

# Copyright (c) 2014, Joshua Stone
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from .request import get_tags, url_tags
from .tag import Tag
from .helpers import tags, join_params, set_limit

__all__ = [
  "Tags"
]

class Tags(object):
  """
  Tags() is the interface for interacting with Derpibooru's API similar on Search().

  All properties are read-only, and every method returns a new instance of
  Tags() to avoid mutating state in ongoing search queries. This makes object
  interactions predictable as well as making versioning of searches relatively
  easy.
  """
  def __init__(self, q=set(), limit=50, per_page=25, page=1,
               url_domain="https://derpibooru.org", proxies={}):
    """
    By default initializes an instance of Tags with the parameters to get
    the first 25 comments on Derpibooru's tags page.
    """
    self.proxies = proxies
    self.url_domain = url_domain
    self._params = {
      "q": tags(q),
      "per_page": set_limit(per_page),
      "page": set_limit(page)
    }
    self._limit = set_limit(limit)
    self._search = get_tags(self._params, self._limit,
                            url_domain=self.url_domain, proxies=self.proxies)
  
  def __iter__(self):
    """
    Make Tags() iterable so that new search results can be lazily generated
    for performance reasons.
    """
    return self

  @property
  def parameters(self):
    """
    Returns a list of available parameters; useful for passing state to new
    instances of Tags().
    """
    return self._params

  @property
  def url(self):
    """
    Returns a search URL built on set parameters. Example based on default
    parameters:

    https://derpibooru.org/tags?tq=%2A
    """
    return url_tags(self._params, url_domain=self.url_domain)

  def query(self, *q):
    """
    Takes one or more strings for searching by query.
    """
    params = join_params(self.parameters, {"q": q,
                                           "limit": self._limit,
                                           "url_domain": self.url_domain,
                                           "proxies": self.proxies}
                        )

    return self.__class__(**params)

  def limit(self, limit):
    """
    Set absolute limit on number of tags to return, or set to None to return
    as many results as needed; default 50 tags. This limit on app-level.
    """
    params = join_params(self.parameters, {"limit": limit,
                                           "url_domain": self.url_domain,
                                           "proxies": self.proxies})

    return self.__class__(**params)

  def query_append(self,*q):
     """
     Adds query to current search.
     """
     query = self._params['q'].union(q)
     params = join_params(self.parameters, {"q": query,
                                           "limit": self._limit,
                                           "url_domain": self.url_domain,
                                           "proxies": self.proxies}
                        )

     return self.__class__(**params)

  def query_remove(self,*q):
     """
     Removes query from current search.
     """
     query = self._params['q'].difference(q)
     params = join_params(self.parameters, {"q": query,
                                           "limit": self._limit,
                                           "url_domain": self.url_domain,
                                           "proxies": self.proxies}
                        )

     return self.__class__(**params)

  def get_page(self,page):
    """
    Set returned page of search.
    """
    params = join_params(self.parameters, {"page": set_limit(page),
                                           "limit": self._limit,
                                           "url_domain": self.url_domain,
                                           "proxies": self.proxies}
                        )

    return self.__class__(**params)

  def per_page(self,limit):
    """
    Set absolute limit on number of tags to get, or set to None to return
    defaulting 25 tags; max 50 tags. This limit on API-level.
    """
    params = join_params(self.parameters, {"per_page": set_limit(limit),
                                           "limit": self._limit,
                                           "url_domain": self.url_domain,
                                           "proxies": self.proxies}
                        )

    return self.__class__(**params)

  def __next__(self):
    """
    Returns a result wrapped in a new instance of Tag().
    """
    return Tag(next(self._search), url_domain=self.url_domain, proxies=self.proxies)



# -*- coding: utf-8 -*-

# Copyright (c) 2014, Joshua Stone
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from requests import get, post, codes
from urllib.parse import urlencode
from .helpers import format_params, format_params_url_galleries, slugging_tag

__all__ = [
  "url", "request", "get_images", "get_image_data", "get_image_faves",
  "url_related", "request_related", "get_related",
  "post_image",
  "url_comments", "request_comments", "get_comments", "get_comment_data",
  "url_tags", "request_tags", "get_tags", "get_tag_data",
  "get_user_id_by_name", "get_user_data",
  "request_filters",
  "get_filters", "get_filter_data",
  "url_galleries", "request_galleries", "get_galleries",
  "request_forums", "get_forums", "get_forum_data",
  "url_topics", "request_topics", "get_topics", "get_topic_data",
  "url_search_posts",
  "url_posts", "request_posts", "get_posts", "get_post_data"
]

def request_content(search, p, items_name, post_request=False, proxies={}):
  if post_request:
    request = post(search, params=p, proxies=proxies)
  else:
    request = get(search, params=p, proxies=proxies)
  if "per_page" not in p:
    p["per_page"] = 50
  while request.status_code == codes.ok:
    items, item_count = request.json()[items_name], 0
    for item in items:
      yield item
      item_count += 1
    if item_count < p["per_page"]:
      break
    p["page"] += 1
    request = get(search, params=p, proxies=proxies)

def get_content(request_func, *request_args, limit=50, **request_kwargs):
  if limit is not None:
    if limit > 0:
      r = request_func(*request_args, **request_kwargs)
      for index, content_item in enumerate(r, start=1):
        yield content_item
        if index >= limit:
          break
  else:
    r = request_func(*request_args, **request_kwargs)
    for content_item in r:
      yield content_item

def url(params, url_domain="https://derpibooru.org"):
  p = format_params(params)
  url = f"{url_domain}/search?{urlencode(p)}"
  return url

def request(params, url_domain="https://derpibooru.org", proxies={}):
  if "reverse_url" in params and params["reverse_url"]:
    search, p = f"{url_domain}/api/v1/json/search/reverse", format_params(params)
    p = {i:p[i] for i in p if i in ('url','distance')}
    post_request = True
  else:
    search, p = f"{url_domain}/api/v1/json/search/images", format_params(params)
    p = {i:p[i] for i in p if i not in ('url','distance')}
    post_request = False
  for image in request_content(search, p, "images", post_request=post_request, proxies=proxies):
    yield image

def get_images(params, limit=50, url_domain="https://derpibooru.org", proxies={}):
  for image in get_content(request, params, limit=limit, url_domain=url_domain, proxies=proxies):
    yield image

def get_image_data(id_number, url_domain="https://derpibooru.org", proxies={}):
  '''id_number can be "featured"'''
  url = f"{url_domain}/api/v1/json/images/{id_number}"

  request = get(url, proxies=proxies)

  if request.status_code == codes.ok:
    data = request.json()

    if data["image"]["duplicate_of"]:
      return get_image_data(data["image"]["duplicate_of"], url_domain=url_domain, proxies=proxies)
    else:
      return data["image"]

def get_image_faves(id_number, url_domain="https://derpibooru.org", proxies={}):
  url = f"{url_domain}/images/{id_number}/favorites"

  request = get(url, proxies=proxies)

  if request.status_code == codes.ok:
    data = request.text.rsplit('</h5>',1)[-1].strip()
    if data.endswith('</a>'):
       data = data[:-4]
    data = data.split("</a> <")
    data = [useritem.rsplit('">',1)[-1] for useritem in data]
    return data

def url_related(id_number, params, url_domain="https://derpibooru.org"):
  p = format_params(params)
  url = f"{url_domain}/images/{id_number}/related?{urlencode(p)}"
  return url

def request_related(id_number, params, url_domain="https://derpibooru.org", proxies={}):
  search, p = f"{url_domain}/images/{id_number}/related", format_params(params)
  request = get(search, params=p, proxies=proxies)

  # It should be temporary solution, until related returns to API
  if request.status_code == codes.ok:
    images = [f"""id:{image.split('"',1)[0]}""" for image in request.text.split('<div class="media-box" data-image-id="')][1:]
  params['q'] = (" || ".join(images),)
  params['sf'] = "_score"
  params['sd'] = "desc"
  search, p = f"{url_domain}/api/v1/json/search/images", format_params(params)

  for image in request_content(search, p, "images", proxies=proxies):
    yield image

def get_related(id_number, params, limit=50, url_domain="https://derpibooru.org", proxies={}):
  for image in get_content(request_related, id_number, params,
                           limit=limit, url_domain=url_domain, proxies=proxies):
    yield image

def post_image(key, image_url, description="", tag_input="", source_url="",
               url_domain="https://derpibooru.org", proxies={}):
  '''
  You must provide the direct link to the image in the image_url parameter.
  Abuse of the endpoint will result in a ban.
  '''
  search = f"{url_domain}/api/v1/json/images"
  json = {"image": {"description": description, 
                    "tag_input": ", ".join(tag_input), 
                    "source_url": source_url
                   },
          "url": image_url
         }
  request = post(search, params={"key": key}, json=json, proxies=proxies)
  if request.status_code == codes.ok:
    data = request.json()
    return data

def url_comments(params, url_domain="https://derpibooru.org"):
  p = format_params(params)
  p["qc"]=p["q"]
  del(p["q"])
  url = f"{url_domain}/comments?{urlencode(p)}"
  return url

def request_comments(params, url_domain="https://derpibooru.org", proxies={}):
  search, p = f"{url_domain}/api/v1/json/search/comments", format_params(params)
  for comment in request_content(search, p, "comments", proxies=proxies):
    yield comment

def get_comments(params, limit=50, url_domain="https://derpibooru.org", proxies={}):
  for comment in get_content(request_comments, params,
                             limit=limit, url_domain=url_domain, proxies=proxies):
    yield comment

def get_comment_data(id_number, url_domain="https://derpibooru.org", proxies={}):
  url = f"{url_domain}/api/v1/json/comments/{id_number}"

  request = get(url, proxies=proxies)

  if request.status_code == codes.ok:
    data = request.json()

    return data["comment"]

def url_tags(params, url_domain="https://derpibooru.org"):
  p = format_params(params)
  p["tq"]=p["q"]
  del(p["q"])
  url = f"{url_domain}/tags?{urlencode(p)}"
  return url

def request_tags(params, url_domain="https://derpibooru.org", proxies={}):
  search, p = f"{url_domain}/api/v1/json/search/tags", format_params(params)
  for tag in request_content(search, p, "tags", proxies=proxies):
    yield tag

def get_tags(params, limit=50, url_domain="https://derpibooru.org", proxies={}):
  for tag in get_content(request_tags, params, limit=limit, url_domain=url_domain, proxies=proxies):
    yield tag

def get_tag_data(tag, url_domain="https://derpibooru.org", proxies={}):
  url = f"{url_domain}/api/v1/json/tags/{tag}"

  request = get(url, proxies=proxies)

  if request.status_code == codes.ok:
    data = request.json()

    return data["tag"]

def get_user_id_by_name(username, url_domain="https://derpibooru.org", proxies={}):
  url = f"{url_domain}/profiles/{slugging_tag(username)}"

  request = get(url, proxies=proxies)

  profile_data = request.text
  user_id = profile_data.split("/conversations?with=",1)[-1].split('">',1)[0]
  return user_id

def get_user_data(user_id, url_domain="https://derpibooru.org", proxies={}):
  url = f"{url_domain}/api/v1/json/profiles/{user_id}"

  request = get(url, proxies=proxies)

  if request.status_code == codes.ok:
    data = request.json()

    return data["user"]

def request_filters(filter_id, params, url_domain="https://derpibooru.org", proxies={}):
  '''filter_id can be "system"'''
  search, p = f"{url_domain}/api/v1/json/filters/{filter_id}", format_params(params)
  for filter_item in request_content(search, p, "filters", proxies=proxies):
    yield filter_item

def get_filters(filter_id, params, url_domain="https://derpibooru.org", limit=50, proxies={}):
  for filter_item in get_content(request_filters, filter_id, params,
                                 limit=limit, url_domain=url_domain, proxies=proxies):
    yield filter_item

def get_filter_data(filter_id, url_domain="https://derpibooru.org", proxies={}):
  url = f"{url_domain}/api/v1/json/filters/{filter_id}"

  request = get(url, proxies=proxies)

  if request.status_code == codes.ok:
    data = request.json()

    return data["filter"]

def url_galleries(params, url_domain="https://derpibooru.org"):
  p = format_params_url_galleries(params)
  url = f"{url_domain}/galleries?{urlencode(p)}"
  return url

def request_galleries(params, url_domain="https://derpibooru.org", proxies={}):
  search, p = f"{url_domain}/api/v1/json/search/galleries", format_params(params)
  for gallery in request_content(search, p, "galleries", proxies=proxies):
    yield gallery

def get_galleries(params, limit=50, url_domain="https://derpibooru.org", proxies={}):
  for gallery in get_content(request_galleries, params, limit=limit, url_domain=url_domain, proxies=proxies):
    yield gallery

def request_forums(params, url_domain="https://derpibooru.org", proxies={}):
  search, p = f"{url_domain}/api/v1/json/forums", format_params(params)
  for forum in request_content(search, p, "forums", proxies=proxies):
    yield forum

def get_forums(params, limit=50, url_domain="https://derpibooru.org", proxies={}):
  for forum in get_content(request_forums, params, limit=limit, url_domain=url_domain, proxies=proxies):
    yield forum

def get_forum_data(short_name, url_domain="https://derpibooru.org", proxies={}):
  url = f"{url_domain}/api/v1/json/forums/{short_name}"

  request = get(url, proxies=proxies)

  if request.status_code == codes.ok:
    data = request.json()
    return data["forum"]

def url_topics(forum_short_name, params, url_domain="https://derpibooru.org"):
  p = format_params(params)
  url = f"{url_domain}/forums/{forum_short_name}?{urlencode(p)}"
  return url

def request_topics(forum_short_name, params, url_domain="https://derpibooru.org", proxies={}):
  search, p = f"{url_domain}/api/v1/json/forums/{forum_short_name}/topics", format_params(params)
  for topic in request_content(search, p, "topics", proxies=proxies):
    yield topic

def get_topics(forum_short_name, params, limit=50, url_domain="https://derpibooru.org", proxies={}):
  for topic in get_content(request_topics, forum_short_name, params,
                           limit=limit, url_domain=url_domain, proxies=proxies):
    yield topic

def get_topic_data(forum_short_name, topic_slug, url_domain="https://derpibooru.org", proxies={}):
  url = f"{url_domain}/api/v1/json/forums/{forum_short_name}/topics/{topic_slug}"

  request = get(url, proxies=proxies)

  if request.status_code == codes.ok:
    data = request.json()
    return data["topic"]

def url_search_posts(params, url_domain="https://derpibooru.org"):
  p = format_params(params)
  url = f"{url_domain}/posts?{urlencode(p)}"
  return url

def url_posts(forum_short_name, topic_slug, params, url_domain="https://derpibooru.org"):
  import math
  p = format_params(params)
  api_page = p['page']
  api_per_page = p['per_page']
  web_per_page = 25
  api_last_post_on_page = api_page * api_per_page
  api_first_post_on_page = api_last_post_on_page - api_per_page + 1 
  p['page'] = math.ceil(api_first_post_on_page / 25)
  del(p['per_page'])
  url = f"{url_domain}/forums/{forum_short_name}/topics/{topic_slug}?{urlencode(p)}"
  return url

def request_posts(params, forum_short_name="", topic_slug="", url_domain="https://derpibooru.org", proxies={}):
  if forum_short_name and topic_slug:
    search, p = f"{url_domain}/api/v1/json/forums/{forum_short_name}/topics/{topic_slug}/posts", format_params(params)
  else:
    search, p = f"{url_domain}/api/v1/json/search/posts", format_params(params)
  for post in request_content(search, p, "posts", proxies=proxies):
    yield post

def get_posts(params, forum_short_name="", topic_slug="", limit=50, url_domain="https://derpibooru.org", proxies={}):
  for post in get_content(request_posts, params,
                          limit=limit, forum_short_name=forum_short_name,
                          topic_slug=topic_slug, url_domain=url_domain, proxies=proxies):
    yield post

def get_post_data(id_number, url_domain="https://derpibooru.org", proxies={}):
  url = f"{url_domain}/api/v1/json/posts/{id_number}"

  request = get(url, proxies=proxies)

  if request.status_code == codes.ok:
    data = request.json()
    return data["post"]